import logging
import queue

textual_log_queue = queue.Queue(maxsize=200)


class TextualLogHandler(logging.Handler):
    """Capture log messages and send them to the log queue."""

    def __init__(self) -> None:
        logging.Handler.__init__(self=self)

    def emit(self, record) -> None:
        textual_log_queue.put(record)



import datetime
import logging

from aprsd import utils
from aprsd.packets import core
from haversine import Unit, haversine
from loguru import logger
from oslo_config import cfg
from rich.text import Text
from textual.app import ComposeResult
from textual.reactive import Reactive
from textual.widget import Widget
from textual.widgets import Label

LOG = logging.getLogger("APRSD")
LOGU = logger
CONF = cfg.CONF
MYCALLSIGN_COLOR = "yellow"


class APRSDPacketWidget(Widget):
    """Display an APRS packet."""

    DEFAULT_CSS = """
    APRSDPacketWidget {
        color: $text;
        width: 100%;
        height: 6;
        margin: 1;
        padding: 0 0 0 0;
    }
    """

    packet: type[core.Packet]
    acked: Reactive[bool] = Reactive(False)

    def __init__(self, packet: type[core.Packet], id: str):
        super().__init__(id=id)
        self.packet = packet

    @property
    def is_tx(self):
        """Did we TX this packet?"""
        return self.packet.from_call == CONF.callsign

    @property
    def from_color(self):
        if self.packet.from_call == CONF.callsign:
            # The packet was sent by us. (TX)
            return f"{MYCALLSIGN_COLOR}"
        else:
            # The packet was sent by someone else. (RX)
            return f"b {utils.hex_from_name(self.packet.from_call)}"

    @property
    def to_color(self):
        if self.packet.from_call == CONF.callsign:
            # The packet was sent by us. (TX)
            return f"b {utils.hex_from_name(self.packet.to_call)}"
        else:
            # The packet was sent by someone else. (RX)
            return f"{MYCALLSIGN_COLOR}"

    def _distance_msg(self):
        # is there distance information?
        if (
            isinstance(self.packet, core.GPSPacket)
            and CONF.latitude
            and CONF.longitude
            and self.packet.latitude
            and self.packet.longitude
        ):
            DEGREES_COLOR = "[b #62C2DD]"
            DEGREES_COLOR_END = "[/b #62C2DD]"
            DISTANCE_COLOR = "[orange]"
            DISTANCE_COLOR_END = "[/orange]"
            my_coords = (float(CONF.latitude), float(CONF.longitude))
            packet_coords = (float(self.packet.latitude), float(self.packet.longitude))
            try:
                bearing = utils.calculate_initial_compass_bearing(
                    my_coords, packet_coords
                )
            except Exception as e:
                LOG.error(f"Failed to calculate bearing: {e}")
                bearing = 0

            # cardinal = utils.degrees_to_cardinal(bearing, full_string=True)
            # cardinal_color = f"b {utils.hex_from_name(cardinal)}"

            return (
                f"{DEGREES_COLOR}{utils.degrees_to_cardinal(bearing, full_string=True)}{DEGREES_COLOR_END} "
                f"{DISTANCE_COLOR}@ {haversine(my_coords, packet_coords, unit=Unit.MILES):.2f}miles{DISTANCE_COLOR_END}"
            )

    def _build_title(self):
        """build the title of the packet."""
        title = []
        from_color = self.from_color
        to_color = self.to_color

        FROM = f"[{from_color}]{self.packet.from_call}[/{from_color}]"
        TO = f"[{to_color}]{self.packet.to_call}[/{to_color}]"
        via_color = "b #1AA730"
        ARROW = f"[{via_color}]\u2192[/{via_color}]"
        title.append(f"{FROM} {ARROW}")
        if self.packet.from_call == CONF.callsign:
            title.append(f"{TO}")
        else:
            title.append(f"{ARROW}".join(self.packet.path))
            title.append(f"{ARROW} {TO}")

        if self.packet.msgNo:
            title.append(f":{self.packet.msgNo}")

        distance_msg = self._distance_msg()
        if distance_msg:
            title.append(f" : {distance_msg}")

        if self.is_tx:
            if self.acked:
                title.append(" 👍")
            else:
                title.append(" 👎")
        self.border_title = " ".join(title)

    def _build_subtitle(self):
        date_str = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        pkt_type = self.packet.__class__.__name__
        pkt_type_color = f"b {utils.hex_from_name(pkt_type)}"
        self.border_subtitle = (
            f"{date_str} [{pkt_type_color}]{pkt_type}[/{pkt_type_color}]"
        )
        self.styles.border_subtitle_color = "rgb(150,150,150)"

    def compose(self) -> ComposeResult:
        self._build_title()
        self._build_subtitle()

        msg_text = Text("", style="bright_white")
        msg_text.append(str(self.packet.human_info))

        raw_header = Text("Raw:", style="grey39")
        raw_text = Text(f"\n{self.packet.raw}", style="grey27")
        msg_text.append("\n\n")
        msg_text.append(raw_header)
        msg_text.append(raw_text)

        yield Label(msg_text)

        if self.packet.from_call == CONF.callsign:
            self.styles.border_title_align = "right"
            self.styles.border = ("solid", "red")
            self.styles.border_subtitle_align = "right"
        else:
            self.styles.border = ("solid", "green")
            self.styles.border_title_align = "left"
            self.styles.border_subtitle_align = "left"



from aprsd.packets import core


def _get_tab_id(tabname: str) -> str:
    """Get the tab id for a tabname."""
    return f"t_{tabname}"


def _get_packet_id(packet: type[core.Packet], timestamp: bool = False) -> str:
    """Get the packet id for a packet."""
    if not packet.msgNo:
        return f"_{hash(packet)}"
    else:
        return f"_{packet.msgNo}"



from oslo_config import cfg


extension_group = cfg.OptGroup(
    name="aprsd_rich_cli_extension",
    title="APRSD rich extension settings",
)

extension_opts = [
    cfg.BoolOpt(
        "enabled",
        default=True,
        help="Enable the extension?",
    ),
]

ALL_OPTS = extension_opts


def register_opts(cfg):
    cfg.register_group(extension_group)
    cfg.register_opts(ALL_OPTS, group=extension_group)


def list_opts():
    return {
        extension_group.name: extension_opts,
    }



import logging

import aprsd
from rich.text import Text
from textual.app import ComposeResult, RenderResult
from textual.containers import Horizontal
from textual.reactive import Reactive

from aprsd_rich_cli_extension.components import packet_widget

LOG = logging.getLogger("APRSD")


class HeaderConnection(Horizontal):
    """Display the title / subtitle in the header."""

    text: Reactive[str] = Reactive("")
    """The main title text."""

    sub_text = Reactive("")
    """The sub-title text."""

    def render(self) -> RenderResult:
        """Render the title and sub-title.

        Returns:
            The value to render.
        """
        text = Text(self.text, no_wrap=True, overflow="ellipsis")
        if self.sub_text:
            text.append(" — ")
            text.append(self.sub_text, packet_widget.MYCALLSIGN_COLOR)
        return text


class HeaderVersion(Horizontal):
    """Display the version in the header."""

    text: Reactive[str] = Reactive("")
    """The main title text."""

    def render(self) -> RenderResult:
        return Text(f"APRSD : {aprsd.__version__}", no_wrap=True, overflow="ellipsis")


class HeaderEarth(Horizontal):
    """The earth icon in the header."""

    def render(self) -> RenderResult:
        # 🌍
        # 🌏
        return Text("🌎 ")


class HeaderFilter(Horizontal):
    """Display the filter in the header."""

    text: Reactive[str] = Reactive("")
    """The main title text."""

    sub_text = Reactive("")
    """The sub-title text."""

    def __init__(self, id: str, filter: str = None):
        super().__init__(id=id)
        LOG.error(f"HeaderFilter: {filter}")
        self.filter = filter
        if self.filter:
            self.text = f"Filter: {self.filter}"
        else:
            self.text = "Filter: None"

    def render(self) -> RenderResult:
        """Render the title and sub-title.

        Returns:
            The value to render.
        """
        text = Text(self.text, no_wrap=True, overflow="ellipsis")
        if self.sub_text:
            text.append(" — ")
            text.append("Num Packets: ")
            text.append(self.sub_text, "yellow")
        return text


class AppHeader(Horizontal):
    """The header of the app."""

    DEFAULT_CSS = """
    SpinnerWidget {
        content-align: left middle;
    }
    """

    def __init__(self, filter: str = None):
        super().__init__()
        self.filter = filter

    def compose(self) -> ComposeResult:
        yield HeaderConnection(id="app-connection")
        earth = HeaderEarth(id="app-earth")
        earth.display = False
        yield earth
        yield HeaderFilter(id="app-filter")
        yield HeaderVersion(id="app-version")



import click

from aprsd.cli_helper import AliasedGroup
from aprsd.main import cli


@cli.group(cls=AliasedGroup, aliases=['rich'], help="APRSD Extension to create textual rich CLI versions of aprsd commands")
@click.pass_context
def rich(ctx):
    pass



"""
Listen to aprs packets and show them.

See https://textual.textualize.io/blog/2024/09/15/anatomy-of-a-textual-user-interface/
"""

# /// script
# requires-python = ">=3.12"
# dependencies = [
#     "aprsd",
#     "textual",
# ]
# ///

import datetime
import logging
import queue
import signal
import time
import typing as t

import click
from aprsd import cli_helper as aprsd_cli_helper
from aprsd import (
    conf,  # noqa: F401
    threads,
)
from aprsd.packets import core
from aprsd.stats import collector
from aprsd.utils import trace
from loguru import logger
from oslo_config import cfg
from textual import work
from textual.app import ComposeResult
from textual.binding import Binding
from textual.containers import Grid, Horizontal, VerticalScroll
from textual.screen import Screen
from textual.widget import Widget
from textual.widgets import Button, Footer, Input, Label, TabbedContent, TabPane

# Import the extension's configuration options
from aprsd_rich_cli_extension import (
    cli_helper,
    cmds,  # noqa
    conf,  # noqa
)
from aprsd_rich_cli_extension.components import (
    base_app,
    header,
    packet_widget,
)
from aprsd_rich_cli_extension.components import utils as components_utils

LOG = logging.getLogger("APRSD")
CONF = cfg.CONF
LOGU = logger
F = t.TypeVar("F", bound=t.Callable[..., t.Any])


@click.version_option()
@click.pass_context
def cli(ctx):
    pass


def signal_handler(sig, frame):
    threads.APRSDThreadList().stop_all()
    if "subprocess" not in str(frame):
        LOG.info(
            "Ctrl+C, Sending all threads exit! Can take up to 10 seconds {}".format(
                datetime.datetime.now(),
            ),
        )
        time.sleep(5)
        # Last save to disk
        collector.Collector().collect()


class PacketStats(Widget):
    def __init__(self):
        self.packet_count = 0
        self.packet_types = {}

    def compose(self) -> ComposeResult:
        yield Label(f"Packet Count: {self.packet_count}")
        yield Label(f"Packet Types: {self.packet_types}")

    def update(self, packet: type[core.Packet]):
        self.packet_count += 1
        self.packet_types[packet.__class__.__name__] = (
            self.packet_types.get(packet.__class__.__name__, 0) + 1
        )


class AppHeader(Horizontal):
    """The header of the app."""

    def __init__(self, filter: str):
        super().__init__()
        self.filter = filter

    def compose(self) -> ComposeResult:
        yield header.HeaderConnection(id="app-connection")
        yield header.HeaderFilter(id="app-filter", filter=self.filter)
        yield header.HeaderVersion(id="app-version")


class APRSFilterInput(Screen):
    CSS = """
        APRSFilterInput {
            align: center middle;
        }

        Grid {
            grid-size: 2 2;
            padding: 0 1;
            width: 40;
            height: 10;
            border: thick $background 80%;
            background: $surface;
        }

        #input_filter {
            column-span: 2;
        }
    """

    def compose(self) -> ComposeResult:
        with Grid():
            yield Input(placeholder="Enter APRS filter", id="input_filter")
            yield Button("Submit", variant="primary", id="submit")
            yield Button("Cancel", variant="error", id="cancel")

    def on_button_pressed(self, event):
        if event.button.id == "submit":
            self.input_text = self.query_one("#input_filter").value
            self.app.filter_changed(self.input_text)
            self.app.pop_screen()


class APRSListenerApp(base_app.APRSDApp):
    """App to display APRS packets in real-time."""

    CSS_PATH = base_app.APRSDApp.CSS_PATH + ["listen.tcss"]

    BINDINGS = base_app.APRSDApp.BINDINGS + [
        Binding(
            "ctrl+f",
            "change_filter",
            "Change Filter",
            tooltip="Set the aprs filter for incoming packets",
        ),
    ]

    filter = None
    pkt_hash = []

    def __init__(self, log_packets: bool = False, filter: str = None):
        self.filter = ",".join(filter)
        super().__init__()

    def do_setup(self):
        self.check_setup()
        self.init_client(filter=self.filter)
        self.chat_binding_count = 1

        self._start_threads()

    def action_change_filter(self):
        self.push_screen(APRSFilterInput(id="aprs-filter-dialog"))

    @work(exclusive=True)
    async def request_filter(self) -> None:
        # stop the client?
        await self.push_screen_wait(APRSFilterInput(id="aprs-filter-dialog"))

    def filter_changed(self, filter: str) -> None:
        LOG.debug(f"Filter_changed to '{filter}'")
        self.filter = filter

    @trace.trace
    def on_mount(self) -> None:
        """Start the APRS listener threads when app starts."""
        # super().on_mount()

        # now add the packet view
        self.query_one("#tabbed-content").add_pane(
            TabPane("Packets", VerticalScroll(id="packet-view"), id="packets-tab")
        )
        self._add_rich_log()
        self.packet_loop()
        self.check_connection()

    def compose(self) -> ComposeResult:
        yield AppHeader(filter=self.filter)
        yield TabbedContent(id="tabbed-content")
        yield Footer()

    @work(exclusive=False)
    async def packet_loop(self) -> None:
        """Check for new packets in a loop."""
        packet_view = self.query_one("#packet-view")
        filter_widget = self.query_one("#app-filter")
        self.packet_count = 0

        while True:
            try:
                # Non-blocking queue check
                packet = base_app.processed_queue.get(timeout=1)
                if packet:
                    self.packet_count += 1
                    filter_widget.sub_text = f"{self.packet_count}"
                    # LOG.error(f"packet_count = {self.packet_count} {hash(packet)}")
                    widget_id = components_utils._get_packet_id(packet)
                    if widget_id not in self.pkt_hash:
                        self.pkt_hash.append(widget_id)
                        found = False
                    else:
                        found = True

                    try:
                        if not found:
                            widget = packet_widget.APRSDPacketWidget(
                                packet, id=widget_id
                            )
                            if widget:
                                await packet_view.mount(widget)
                    except Exception as e:
                        LOG.error(f"Error mounting packet: {e}")
                        self.notify(f"Error adding packet: {packet}")

                    # Scroll to bottom
                    packet_view.scroll_end(animate=False)
                    if len(packet_view.children) > 10:
                        Widget.remove(packet_view.children[0])
                        self.pkt_hash.remove(self.pkt_hash[0])
            except queue.Empty:
                # No packets, wait a bit
                # await asyncio.sleep(0.05)
                pass
        LOG.error("check_packets: done")


@cmds.rich.command()
@aprsd_cli_helper.add_options(aprsd_cli_helper.common_options)
@click.option("--log-packets", is_flag=True, help="Log packets to the console.")
@click.argument(
    "filter",
    nargs=-1,
    required=True,
)
@click.pass_context
@cli_helper.process_standard_options
def listen(ctx, log_packets: bool, filter: str):
    """Listen to APRS packets in the terminal."""
    signal.signal(signal.SIGINT, signal_handler)
    signal.signal(signal.SIGTERM, signal_handler)
    app = APRSListenerApp(log_packets=log_packets, filter=filter)
    app.run()



from textual.app import ComposeResult
from textual.containers import Grid
from textual.screen import ModalScreen
from textual.widgets import Button, Input


class AddChatScreen(ModalScreen[str]):
    """The screen to add a new chat."""

    CSS = """
        AddChatScreen {
            align: center middle;
        }

        Grid {
            grid-size: 2 2;
            padding: 0 1;
            width: 40;
            height: 10;
            border: thick $background 80%;
            background: $surface;
        }

        #input_callsign {
            column-span: 2;
        }
    """

    def compose(self) -> ComposeResult:
        with Grid():
            yield Input(
                placeholder="Enter callsign", id="input_callsign", max_length=60
            )
            yield Button("Add", id="submit")
            yield Button("Cancel", id="cancel")

    def on_button_pressed(self, event):
        if event.button.id == "submit":
            input_text = self.query_one("#input_callsign").value
            self.dismiss(input_text)



# Override the aprsd.cli_helper.process_standard_options function
# to allows us to inject a custom log handler.
import logging
import typing as t
from functools import update_wrapper

import aprsd
from aprsd.log import log
from aprsd.utils import trace
from oslo_config import cfg

from aprsd_rich_cli_extension.log import TextualLogHandler

CONF = cfg.CONF
F = t.TypeVar("F", bound=t.Callable[..., t.Any])


def process_standard_options(f: F) -> F:
    def new_func(*args, **kwargs):
        ctx = args[0]
        ctx.ensure_object(dict)
        config_file_found = True
        if kwargs["config_file"]:
            default_config_files = [kwargs["config_file"]]
        else:
            default_config_files = None

        try:
            CONF(
                [],
                project="aprsd",
                version=aprsd.__version__,
                default_config_files=default_config_files,
            )
        except cfg.ConfigFilesNotFoundError:
            config_file_found = False
        ctx.obj["loglevel"] = kwargs["loglevel"]
        # ctx.obj["config_file"] = kwargs["config_file"]
        ctx.obj["quiet"] = kwargs["quiet"]

        # Create the custom log handler entry
        custom_handler = {
            "sink": TextualLogHandler(),
            "serialize": False,
            "format": CONF.logging.logformat,
            "colorize": False,
            "level": ctx.obj["loglevel"],
        }
        log.setup_logging(
            ctx.obj["loglevel"],
            quiet=True,
            custom_handler=custom_handler,
        )
        if CONF.trace_enabled:
            trace.setup_tracing(["method", "api"])

        if not config_file_found:
            LOG = logging.getLogger("APRSD")  # noqa: N806
            LOG.error("No config file found!! run 'aprsd sample-config'")

        del kwargs["loglevel"]
        del kwargs["config_file"]
        del kwargs["quiet"]
        return f(*args, **kwargs)

    return update_wrapper(t.cast(F, new_func), f)



# this is the base class for all the TUI apps
# for APRSD.
# It provides the same functionality for all the apps.
# It provides a tabbed interface, a log, a status bar, and a help screen.

import asyncio
import logging
import queue
import sys
import typing as t

from aprsd import client as aprsd_client
from aprsd import (
    conf,  # noqa: F401
    threads,
)
from aprsd.client.client import APRSDClient
from aprsd.packets import core
from aprsd.threads import keepalive, rx, service
from aprsd.utils import trace
from loguru import logger
from oslo_config import cfg
from textual import work
from textual.app import App, ComposeResult
from textual.binding import Binding
from textual.widgets import (
    Footer,
    RichLog,
    TabbedContent,
    TabPane,
)

from aprsd_rich_cli_extension import log as log_extension
from aprsd_rich_cli_extension.components import header, help_screen
from aprsd_rich_cli_extension.components import utils as component_utils

LOG = logging.getLogger("APRSD")
CONF = cfg.CONF
LOGU = logger
F = t.TypeVar("F", bound=t.Callable[..., t.Any])


processed_queue = queue.Queue(maxsize=500)


class APRSDListenProcessThread(rx.APRSDFilterThread):
    """Thread that processes packets received from the APRS Client."""

    count = 0

    def __init__(
        self,
        packet_queue,
        processed_queue,
        log_packets=False,
    ):
        super().__init__("ListenProcThread", packet_queue)
        self.processed_queue = processed_queue
        self.log_packets = True

    def print_packet(self, packet):
        if self.log_packets:
            # packet_log.log(packet)
            pass

    def process_packet(self, packet: type[core.Packet]):
        self.processed_queue.put(packet)
        self.count += 1


class APRSDApp(App):
    """App to allow APRS chat in the terminal."""

    CSS_PATH = ["global.tcss"]

    BINDINGS = [
        Binding(
            "ctrl+d",
            "toggle_dark",
            "Toggle Dark",
            tooltip="Switch between light and dark themes",
        ),
        Binding(
            "f1",
            "show_help",
            "Show Help",
            tooltip="Show the help screen",
        ),
        Binding(
            "ctrl+l",
            "show_log",
            "Show Log",
            tooltip="Show the log screen",
        ),
    ]

    def __init__(self):
        super().__init__()
        # Create the queues to be used later by
        # the threads
        # the queue used to process packets that are received
        # from the APRS network
        self.processed_queue = processed_queue
        self.do_setup()

    def do_setup(self):
        self.check_setup()
        self.init_client()
        self.chat_binding_count = 1

        self._start_threads()

    @trace.trace
    def on_mount(self) -> None:
        """Called when the app is started and ready to go."""
        # Start checking for packets
        # self.packet_loop()
        # self.check_connection()
        # self._add_rich_log()

    def check_setup(self):
        # Initialize the client factory and create
        # The correct client object ready for use
        if not APRSDClient().is_enabled:
            LOG.error("No Clients are enabled in config.")
            sys.exit(-1)

        # Make sure we have 1 client transport enabled
        if not APRSDClient().is_configured:
            LOG.error("APRS client is not properly configured in config file.")
            sys.exit(-1)

    def init_client(self, filter: str = None):
        # Creates the client object
        LOG.info("Creating client connection")
        self.aprs_client = APRSDClient()
        LOG.info(self.aprs_client)
        if not self.aprs_client.login_success:
            # We failed to login, will just quit!
            msg = f"Login Failure: {self.aprs_client.login_failure}"
            LOG.error(msg)
            print(msg)
            sys.exit(-1)

        if filter:
            LOG.debug(f"Filter messages on aprsis server by '{filter}'")
            self.aprs_client.set_filter(filter)
        else:
            LOG.debug("No filter set")

    def _start_threads(self):
        # thread to receive packets from the APRS network
        self.listen_thread = rx.APRSDRXThread(
            packet_queue=threads.packet_queue,
        )
        # thread to process packets received from listen_thread
        self.process_thread = APRSDListenProcessThread(
            packet_queue=threads.packet_queue,
            processed_queue=self.processed_queue,
        )
        service.ServiceThreads().register(self.listen_thread)
        service.ServiceThreads().register(self.process_thread)
        service.ServiceThreads().register(keepalive.KeepAliveThread())
        service.ServiceThreads().start()

    @trace.trace
    def _add_rich_log(self):
        """Add a rich log to the app."""
        tabbed_content = self.query_one(TabbedContent)
        tab_id = component_utils._get_tab_id("log")
        tabbed_content.add_pane(
            TabPane(
                "Log",
                RichLog(id="log-scroll"),
                id=tab_id,
            )
        )
        tabbed_content.active = tab_id
        # now start the worker to process the log queue
        self.process_log_queue()

    def action_show_log(self):
        """Show the log screen."""
        tabbed_content = self.query_one(TabbedContent)
        tabbed_content.active = component_utils._get_tab_id("log")

    def action_show_help(self):
        """Show the help screen."""
        self.notify("Showing help screen")
        self.push_screen(help_screen.HelpScreen())

    def action_show_tab(self, tab_id: str):
        """Show a tab."""
        self.query_one(TabbedContent).active = tab_id

    def compose(self) -> ComposeResult:
        yield header.AppHeader()
        yield TabbedContent(id="tabbed-content")
        yield Footer()

    # def on_key(self, event: events.Key) -> None:
    #    self.notify(f"Key pressed: {event}")
    # self.query_one(RichLog).write(event)

    def on_unmount(self) -> None:
        """Stop threads when app exits."""
        threads.APRSDThreadList().stop_all()

    @work(exclusive=False)
    async def process_log_queue(self) -> None:
        """Process the log queue."""
        self.notify("Log queue processing started")
        while True:
            try:
                record = log_extension.textual_log_queue.get_nowait()
                rich_log = self.query_one("#log-scroll")
                rich_log.write(record.getMessage(), expand=True)
            except queue.Empty:
                await asyncio.sleep(0.1)
            except Exception:
                await asyncio.sleep(0.1)

        self.notify("Log queue processing stopped")

    @work(exclusive=False)
    async def packet_loop(self) -> None:
        """Process packets in a loop."""
        self.packet_count = 0

        while True:
            try:
                # Non-blocking queue check
                packet = self.processed_queue.get_nowait()
                self.packet_count += 1

                # Child class will handle the packet
                await self.process_packet(packet)

            except queue.Empty:
                # No packets, wait a bit
                await asyncio.sleep(0.1)

    @work(exclusive=False)
    async def _update_earth(self):
        """Show the earth icon for 2 seconds when we send a beacon."""
        earth = self.query_one("#app-earth")
        earth.display = not earth.display
        await asyncio.sleep(2)
        earth.display = not earth.display

    def _build_connection_string(self, stats) -> str:
        match stats["transport"]:
            case aprsd_client.TRANSPORT_APRSIS:
                transport_name = "APRS-IS"
                connection_string = f"{transport_name} : {stats['server_string']}"
            case aprsd_client.TRANSPORT_TCPKISS:
                transport_name = "TCP/KISS"
                connection_string = (
                    f"{transport_name} : {CONF.kiss_tcp.host}:{CONF.kiss_tcp.port}"
                )
            case aprsd_client.TRANSPORT_SERIALKISS:
                transport_name = "Serial/KISS"
                connection_string = f"{transport_name} : {CONF.kiss_serial.device}"
        return connection_string

    @work(exclusive=False)
    async def check_connection(self) -> None:
        """Check for connection to APRS server."""
        while True:
            if self.aprs_client:
                stats = self.aprs_client.stats()
            try:
                connection_widget = self.query_one("#app-connection")
                connection_string = self._build_connection_string(stats)
                sub_text = CONF.callsign
                if not self.listen_thread.is_alive():
                    connection_widget.text = "Connection Lost"
                    connection_widget.sub_text = ""
                else:
                    connection_widget.text = f"{connection_string}"
                    connection_widget.sub_text = f"{sub_text}"
            except Exception as e:
                LOG.error(f"check_connection: error: {e}")
                await asyncio.sleep(1)

            await asyncio.sleep(1)



"""
APRS Chat for the terminal!

See https://textual.textualize.io/blog/2024/09/15/anatomy-of-a-textual-user-interface/
"""

# /// script
# requires-python = ">=3.12"
# dependencies = [
#     "aprsd",
#     "textual",
# ]
# ///

import asyncio
import datetime
import logging
import queue
import signal
import sys
import time
import typing as t

import aprsd
import click
from aprsd import cli_helper as aprsd_cli_helper
from aprsd import client as aprsd_client
from aprsd import (
    conf,  # noqa: F401
    threads,
)
from aprsd.client.client import APRSDClient
from aprsd.packets import core
from aprsd.stats import collector
from aprsd.threads import aprsd as aprsd_threads
from aprsd.threads import keepalive, rx, service, tx
from loguru import logger
from oslo_config import cfg
from rich.text import Text
from textual import on, work
from textual.app import App, ComposeResult, RenderResult
from textual.binding import Binding
from textual.containers import Horizontal, VerticalScroll
from textual.reactive import Reactive
from textual.widget import Widget
from textual.widgets import (
    Footer,
    Input,
    RichLog,
    TabbedContent,
    TabPane,
)

# Import the extension's configuration options
from aprsd_rich_cli_extension import (
    cli_helper,
    cmds,  # noqa
)
from aprsd_rich_cli_extension import log as log_extension
from aprsd_rich_cli_extension.components import (
    add_chat_screen,
    help_screen,
    packet_widget,
)
from aprsd_rich_cli_extension.components import utils as components_utils

LOG = logging.getLogger('APRSD')
CONF = cfg.CONF
LOGU = logger
F = t.TypeVar('F', bound=t.Callable[..., t.Any])


@click.version_option()
@click.pass_context
def cli(ctx):
    pass


def signal_handler(sig, frame):
    collector.Collector().stop_all()
    threads.APRSDThreadList().stop_all()
    if 'subprocess' not in str(frame):
        LOG.info(
            'Ctrl+C, Sending all threads exit! Can take up to 10 seconds {}'.format(
                datetime.datetime.now(),
            ),
        )
        time.sleep(5)
        # Last save to disk
        collector.Collector().collect()


def _get_scroll_id(callsign: str) -> str:
    """Get the scroll id for a callsign."""
    return f'{callsign}-scroll'


def _get_tab_id(callsign: str) -> str:
    """Get the tab id for a callsign."""
    return f'tab-{callsign}'


class APRSDListenProcessThread(rx.APRSDProcessPacketThread):
    def __init__(
        self,
        packet_queue,
        processed_queue,
    ):
        super().__init__(packet_queue=packet_queue)
        # The processed queue are packets that need to be displayed
        # in the UI
        self.processed_queue = processed_queue

    def process_ack_packet(self, packet):
        """We got an ack for a message, no need to resend it."""
        super().process_ack_packet(packet)
        self.processed_queue.put(packet)

    def process_our_message_packet(self, packet: type[core.Packet]):
        """Process a packet and add it to the processed queue."""
        self.processed_queue.put(packet)


class APRSTXThread(aprsd_threads.APRSDThread):
    """Thread to pull messages from the queue and send them to the APRS server.

    We have to do this to allow the UI to update while the thread is sending messages.

    """

    def __init__(self, packet_queue):
        super().__init__('APRSTXThread')
        self.tx_queue = packet_queue

    def loop(self):
        """Process a packet and add it to the processed queue."""
        while not self.thread_stop:
            if not self.tx_queue.empty():
                packet = self.tx_queue.get()
                tx.send(packet)
            else:
                time.sleep(0.1)


class MyBeaconSendThread(aprsd_threads.APRSDThread):
    """Thread that sends a GPS beacon packet periodically.

    Settings are in the [DEFAULT] section of the config file.
    """

    _loop_cnt: int = 1

    def __init__(self, notify_queue):
        super().__init__('BeaconSendThread')
        self._loop_cnt = 1
        self.notify_queue = notify_queue
        # Make sure Latitude and Longitude are set.
        if not CONF.latitude or not CONF.longitude:
            LOG.error(
                'Latitude and Longitude are not set in the config file.'
                'Beacon will not be sent and thread is STOPPED.',
            )
            self.stop()
        LOG.info(
            'Beacon thread is running and will send '
            f'beacons every {CONF.beacon_interval} seconds.',
        )

    def loop(self):
        # Only dump out the stats every N seconds
        if self._loop_cnt % CONF.beacon_interval == 0:
            pkt = core.BeaconPacket(
                from_call=CONF.callsign,
                to_call='APRS',
                latitude=float(CONF.latitude),
                longitude=float(CONF.longitude),
                comment='APRSD GPS Beacon',
                symbol=CONF.beacon_symbol,
            )
            try:
                # Only send it once
                pkt.retry_count = 1
                tx.send(pkt, direct=True)
                self.notify_queue.put('Beacon sent')
            except Exception as e:
                LOG.error(f'Failed to send beacon: {e}')
                APRSDClient().reset()
                time.sleep(5)

        self._loop_cnt += 1
        time.sleep(1)
        return True


# All of the Textual Widgets needed for the UI


class HeaderConnection(Horizontal):
    """Display the title / subtitle in the header."""

    text: Reactive[str] = Reactive('')
    """The main title text."""

    sub_text = Reactive('')
    """The sub-title text."""

    def render(self) -> RenderResult:
        """Render the title and sub-title.

        Returns:
            The value to render.
        """
        text = Text(self.text, no_wrap=True, overflow='ellipsis')
        if self.sub_text:
            text.append(' — ')
            text.append(self.sub_text, packet_widget.MYCALLSIGN_COLOR)
        return text


class HeaderVersion(Horizontal):
    """Display the version in the header."""

    text: Reactive[str] = Reactive('')
    """The main title text."""

    def render(self) -> RenderResult:
        return Text(f'APRSD : {aprsd.__version__}', no_wrap=True, overflow='ellipsis')


class HeaderEarth(Horizontal):
    """The earth icon in the header."""

    def render(self) -> RenderResult:
        # 🌍
        # 🌏
        return Text('🌎 ')


class AppHeader(Horizontal):
    """The header of the app."""

    DEFAULT_CSS = """
    SpinnerWidget {
        content-align: left middle;
    }
    """

    def __init__(self):
        super().__init__()

    def compose(self) -> ComposeResult:
        yield HeaderConnection(id='app-connection')
        earth = HeaderEarth(id='app-earth')
        earth.display = False
        yield earth
        yield HeaderVersion(id='app-version')


class ChatInput(Horizontal):
    """The input for the chat."""

    DEFAULT_CSS = """
    ChatInput {
        dock: bottom;
        height: 3;
        width: 100%;
        margin-bottom: 1;
        background: $panel;
    }
    Input {
        align: left middle;
        width: 95%;
    }
    """

    @on(Input.Submitted)
    def on_input_submitted(self, event: Input.Submitted) -> None:
        """Handle the input submitted event."""
        LOG.info(f'Input submitted: {event.value}')
        msg_text = event.value
        self.app.action_send_message(msg_text)
        self.query_one('#message-input').value = ''

    def compose(self) -> ComposeResult:
        yield Input(placeholder='Enter message', id='message-input')


class APRSChatApp(App):
    """App to allow APRS chat in the terminal."""

    CSS_PATH = ['global.tcss', 'chat.tcss']

    BINDINGS = [
        Binding(
            'ctrl+q',
            'quit',
            'Quit',
            tooltip='Quit the app',
        ),
        Binding(
            'ctrl+n',
            'add_new_chat',
            'Add New Chat',
            tooltip='Add a chat with a new callsign',
        ),
        Binding(
            'ctrl+d',
            'toggle_dark',
            'Toggle Dark',
            tooltip='Switch between light and dark themes',
        ),
        Binding(
            'ctrl+l',
            'show_log',
            'Show Log',
            tooltip='Show the log screen',
        ),
        Binding(
            'f1',
            'show_help',
            'Show Help',
            tooltip='Show the help screen',
        ),
    ]

    def __init__(self):
        super().__init__()
        # Create the queues to be used later by
        # the threads
        # the queue used to process packets that are received
        # from the APRS network
        self.processed_queue = queue.Queue()
        # the queue used to send packets to the APRS network
        self.tx_queue = queue.Queue()
        # the queue used to notify the user that a beacon has been sent
        self.beacon_notify_queue = queue.Queue()

        self.callsign_tabs = {}

    def on_mount(self) -> None:
        """Called when the app is started and ready to go."""
        # Show the initial splash screen.
        self.check_setup()
        self.init_client()
        self.chat_binding_count = 1

        # packets to be sent to the UI
        self.listen_thread = rx.APRSDRXThread(
            packet_queue=threads.packet_queue,
        )
        self.process_thread = APRSDListenProcessThread(
            packet_queue=threads.packet_queue,
            processed_queue=self.processed_queue,
        )
        self.tx_thread = APRSTXThread(
            packet_queue=self.tx_queue,
        )

        self._start_threads()

        # Start checking for packets
        self.process_packets()
        self.check_connection()
        if CONF.enable_beacon:
            self.check_beacon_notify()

        self._add_rich_log()

    def _add_rich_log(self):
        """Add a rich log to the app."""
        tabbed_content = self.query_one(TabbedContent)
        tab_id = _get_tab_id('log')
        tabbed_content.add_pane(
            TabPane(
                'Log',
                RichLog(id='log-scroll'),
                id=tab_id,
            )
        )
        tabbed_content.active = tab_id
        # now start the worker to process the log queue
        self.process_log_queue()

    def action_show_log(self):
        """Show the log screen."""
        tabbed_content = self.query_one(TabbedContent)
        tabbed_content.active = _get_tab_id('log')

    def check_setup(self):
        # Initialize the client factory and create
        # The correct client object ready for use
        if not APRSDClient().is_enabled:
            LOG.error('No Clients are enabled in config.')
            sys.exit(-1)

        # Make sure we have 1 client transport enabled
        if not APRSDClient().is_configured:
            LOG.error('APRS client is not properly configured in config file.')
            sys.exit(-1)

    def init_client(self):
        # Creates the client object
        LOG.info('Creating client connection')
        self.aprs_client = APRSDClient()
        LOG.info(self.aprs_client)
        if not self.aprs_client.login_success:
            # We failed to login, will just quit!
            msg = f'Login Failure: {self.aprs_client.login_failure}'
            LOG.error(msg)
            print(msg)
            sys.exit(-1)

    def _start_threads(self):
        service.ServiceThreads().register(self.listen_thread)
        service.ServiceThreads().register(self.process_thread)
        service.ServiceThreads().register(self.tx_thread)
        service.ServiceThreads().register(keepalive.KeepAliveThread())
        if CONF.enable_beacon:
            LOG.info('Beacon Enabled.  Starting Beacon thread.')
            service.ServiceThreads().register(
                MyBeaconSendThread(self.beacon_notify_queue)
            )
        service.ServiceThreads().start()

    def _get_active_callsign(self):
        """Get the active callsign from the active tab."""
        active_tab = self.query_one(TabbedContent).active
        return str(active_tab).replace('tab-', '')

    def _get_scroll_for_callsign(self, callsign: str):
        """Get the scroll view for a callsign."""
        try:
            scroll = self.query_one(f'#{_get_scroll_id(callsign)}')
            return scroll
        except Exception as e:
            LOG.error(f'Error getting scroll for callsign {callsign}: {e}')
            return None

    def _get_tab_for_callsign(self, callsign: str):
        """Get the tab for a callsign."""
        if callsign in self.callsign_tabs:
            return self.callsign_tabs[callsign]

        try:
            tab = self.query_one(f'#{_get_tab_id(callsign)}')
            return tab
        except Exception as e:
            LOG.error(f'Error getting tab for callsign {callsign}: {e}')
            return None

    def action_add_new_chat(self):
        """When the user asks to create a chat with a new callsign."""
        self.push_screen(add_chat_screen.AddChatScreen(), callback=self._on_add_chat)

    def action_show_help(self):
        """Show the help screen."""
        self.notify('Showing help screen')
        self.push_screen(help_screen.HelpScreen())

    def action_send_message(self, msg_text: str):
        """Send a message to the APRS server."""
        # Get the active callsign
        active_callsign = self._get_active_callsign()
        # self.notify(f"Sending message '{msg_text}' to {active_callsign}")

        # Create the message packet
        msg = core.MessagePacket(
            from_call=CONF.callsign,
            to_call=active_callsign,
            message_text=msg_text,
        )
        msg.prepare(create_msg_number=True)
        self.processed_queue.put(msg)
        self.tx_queue.put(msg)

    def action_show_tab(self, tab_id: str):
        """Show a tab."""
        self.query_one(TabbedContent).active = tab_id

    def _on_add_chat(self, callsign: str) -> None:
        """Handle the result of the add chat screen."""
        callsign = callsign.strip().upper()
        self.notify(f'Adding new chat with callsign: {callsign}')
        # self.notify(f"Adding new chat with callsign: {callsign}")
        if callsign:
            LOG.info(f'Adding new chat with callsign: {callsign}')
            # get the tabbedcontent and add a new pane
            tabbed_content = self.query_one('#tabbed-content')
            tab_id = _get_tab_id(callsign)
            log_pane = self.query_one(f'#{_get_tab_id("log")}')
            tabbed_content.add_pane(
                TabPane(
                    callsign,
                    VerticalScroll(id=_get_scroll_id(callsign)),
                    id=tab_id,
                ),
                before=log_pane,
            )
            self.chat_binding_count += 1
            self.bind(
                # Bind the chat tab to an F key
                f'f{self.chat_binding_count}',
                f"show_tab('{tab_id}')",
                description=f'{callsign}',
            )
            # set the new tab to be active
            tabbed_content.active = tab_id
            self.callsign_tabs[callsign] = self.query_one(f'#{tab_id}')

        # set the focus on the input
        self.query_one('#message-input').focus()

    def compose(self) -> ComposeResult:
        yield AppHeader()
        yield TabbedContent(id='tabbed-content')
        yield ChatInput()
        yield Footer()

    # def on_key(self, event: events.Key) -> None:
    #    self.notify(f"Key pressed: {event}")
    # self.query_one(RichLog).write(event)

    def on_unmount(self) -> None:
        """Stop threads when app exits."""
        threads.APRSDThreadList().stop_all()

    @work(exclusive=False)
    async def process_log_queue(self) -> None:
        """Process the log queue."""
        while True:
            try:
                record = log_extension.textual_log_queue.get_nowait()
                rich_log = self.query_one('#log-scroll')
                rich_log.write(record.getMessage(), expand=True)
            except queue.Empty:
                await asyncio.sleep(0.1)
            except Exception:
                await asyncio.sleep(0.1)

        self.notify('Log queue processing stopped')

    @work(exclusive=False)
    async def process_packets(self) -> None:
        """Process packets in a loop."""
        self.packet_count = 0

        while True:
            try:
                # Non-blocking queue check
                packet = self.processed_queue.get_nowait()
                self.packet_count += 1

                callsign = packet.from_call
                if packet.from_call == CONF.callsign:
                    # this is a message we sent.
                    callsign = packet.to_call
                else:
                    # make sure there is a tab existing for this callsign
                    if not self._get_tab_for_callsign(callsign):
                        self._on_add_chat(callsign)

                if isinstance(packet, core.AckPacket):
                    try:
                        pkt_widget = self.query_one(
                            f'#{components_utils._get_packet_id(packet)}'
                        )
                        if pkt_widget:
                            pkt_widget.acked = True
                            pkt_widget.refresh(recompose=True)
                    except Exception as e:
                        LOG.error(f'Error getting packet widget: {e}')

                scroll_view = self._get_scroll_for_callsign(callsign)
                if scroll_view:
                    if isinstance(packet, core.MessagePacket):
                        await scroll_view.mount(
                            packet_widget.APRSDPacketWidget(
                                packet, id=components_utils._get_packet_id(packet)
                            )
                        )
                        # self.notify(f"Packet({packet.from_call}): '{packet.message_text}' {scroll_view}")
                        # Scroll to bottom
                        scroll_view.scroll_end(animate=False)
                        if len(scroll_view.children) > 10:
                            Widget.remove(scroll_view.children[0])

                        if self._get_active_callsign() != callsign:
                            self.notify(f'New message from {callsign}')
                            # Can we change the color of the tab to red?
                            tab_id = _get_tab_id(callsign)
                            tab = self.query_one(f'#{tab_id}')
                            if tab:
                                tab.styles.background = 'red'
                                self.notify(f'Tab: {tab.classes}')
                                # tab.refresh(recompose=True)
                            ass = self.query(f'#{callsign}')
                            for a in ass:
                                self.notify(f'Child: {a}')
                else:
                    # put the packet back in the queue
                    # the scroll view is not found, so we need to wait a bit
                    # and try again
                    self.processed_queue.put(packet)
                    await asyncio.sleep(0.2)

            except queue.Empty:
                # No packets, wait a bit
                await asyncio.sleep(0.1)

    @work(exclusive=False)
    async def check_beacon_notify(self):
        while True:
            try:
                _ = self.beacon_notify_queue.get_nowait()
                self._update_earth()
            except queue.Empty:
                await asyncio.sleep(1)

    @work(exclusive=False)
    async def _update_earth(self):
        """Show the earth icon for 2 seconds when we send a beacon."""
        earth = self.query_one('#app-earth')
        earth.display = not earth.display
        await asyncio.sleep(2)
        earth.display = not earth.display

    def _build_connection_string(self, stats) -> str:
        match stats['transport']:
            case aprsd_client.TRANSPORT_APRSIS:
                transport_name = 'APRS-IS'
                connection_string = f'{transport_name} : {stats["server_string"]}'
            case aprsd_client.TRANSPORT_TCPKISS:
                transport_name = 'TCP/KISS'
                connection_string = (
                    f'{transport_name} : {CONF.kiss_tcp.host}:{CONF.kiss_tcp.port}'
                )
            case aprsd_client.TRANSPORT_SERIALKISS:
                transport_name = 'Serial/KISS'
                connection_string = f'{transport_name} : {CONF.kiss_serial.device}'
        return connection_string

    @work(exclusive=False)
    async def check_connection(self) -> None:
        """Check for connection to APRS server."""
        while True:
            if self.aprs_client:
                stats = self.aprs_client.stats()
            try:
                connection_widget = self.query_one('#app-connection')
                connection_string = self._build_connection_string(stats)
                sub_text = CONF.callsign
                if not self.listen_thread.is_alive():
                    connection_widget.text = 'Connection Lost'
                    connection_widget.sub_text = ''
                else:
                    connection_widget.text = f'{connection_string}'
                    connection_widget.sub_text = f'{sub_text}'
            except Exception as e:
                LOG.error(f'check_connection: error: {e}')
                await asyncio.sleep(1)

            await asyncio.sleep(1)


@cmds.rich.command()
@aprsd_cli_helper.add_options(aprsd_cli_helper.common_options)
@click.pass_context
@cli_helper.process_standard_options
def chat(ctx):
    """APRS Chat in the terminal."""
    signal.signal(signal.SIGINT, signal_handler)
    signal.signal(signal.SIGTERM, signal_handler)
    tui = APRSChatApp()
    tui.run()
